/*
 * ESLinq - LINQ for EcmaScript 2015
 *
 * An easy and elegant way of working with iterables.
 *
 * Full source code and examples:
 * http://github.com/balazsbotond/eslinq
 */

/* 
 * The MIT License (MIT)
 * 
 * Copyright (c) 2015 Botond Balazs <balazsbotond@gmail.com>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

"use strict";

/**
 * Helper function that wraps the specified iterable instance in an
 * ESLinq Sequence which can be queried using ESLinq operators (like
 * 'select', 'where', etc.).
 * 
 * @param {Iterable} iterable An iterable object (like an Array, a Set,
 *     or any object with a [Symbol.iterator] property).
 * @return {Sequence} An ESLinq Sequence which can be queried using
 *     ESLinq operators (like 'select', 'where', etc.).
 * 
 * @example
 * const numbers = [1, 2, 3, 4, 5];
 * const squaresOfEvenNumbers =
 *     from(numbers)
 *         .where(n => n % 2 === 0)
 *         .select(n => n * n);
 * 
 * for (let n of squaresOfEvenNumbers)
 *     console.log(n); // 4 16
 */

var _createClass = require("babel-runtime/helpers/create-class")["default"];

var _classCallCheck = require("babel-runtime/helpers/class-call-check")["default"];

var _defineProperty = require("babel-runtime/helpers/define-property")["default"];

var _Symbol$iterator = require("babel-runtime/core-js/symbol/iterator")["default"];

var _regeneratorRuntime = require("babel-runtime/regenerator")["default"];

var _getIterator = require("babel-runtime/core-js/get-iterator")["default"];

var _Set = require("babel-runtime/core-js/set")["default"];

var _Array$from = require("babel-runtime/core-js/array/from")["default"];

var _Map = require("babel-runtime/core-js/map")["default"];

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports["default"] = from;

function from(iterable) {
    return new Sequence(iterable);
}

/**
 * An iterable sequence that can be queried using ESLinq operators (like
 * 'select', 'where', etc.).
 * 
 * @implements {Iterable}
 */

var Sequence = (function () {
    function Sequence(iterable) {
        _classCallCheck(this, Sequence);

        this.iterable = iterable;
        this[_Symbol$iterator] = iterable[_Symbol$iterator];
    }

    /*
     * Functions used as default arguments
     */

    /*
     * Projection and restriction methods
     */

    /**
     * Applies the specified transformation to all elements of the
     * `Sequence`, returning a new `Sequence` of the transformed elements.
     * 
     * @param {function(item: *, index: number): *} transform A function
     *     that is used to transform the elements of the sequence. The first
     *     argument, `item`, is the current sequence element, the second one,
     *     `index`, is the zero-based index of the current element.
     * 
     * @return {Sequence} A new `Sequence` of the transformed elements.
     * 
     * @throws {TypeError} if `transform` is not a function
     * 
     * @example
     * // Simple case, only the `item` parameter of `matches` is used
     * const numbers = [1, 2, 3, 4, 5],
     *       even = from(numbers).select(n => n % 2 === 0);
     * 
     * for (let n of even) {
     *     console.log(n); // 2 4
     * }
     * 
     * // Here we also use the `index` parameter
     * const numbers = [1, 2, 3, 4, 5],
     *       numbersPlusIndices = from(numbers).select((n, i) => n + i);
     * 
     * for (let n of numbersPlusIndices) {
     *     console.log(n); // 1 3 5 7 9
     * }
     */

    _createClass(Sequence, [{
        key: "select",
        value: function select(transform) {
            ensureIsFunction(transform, "`transform` should be a function");

            var generator = _regeneratorRuntime.mark(function generator() {
                var index, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, item;

                return _regeneratorRuntime.wrap(function generator$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            index = 0;
                            _iteratorNormalCompletion = true;
                            _didIteratorError = false;
                            _iteratorError = undefined;
                            context$3$0.prev = 4;
                            _iterator = _getIterator(this.iterable);

                        case 6:
                            if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                                context$3$0.next = 14;
                                break;
                            }

                            item = _step.value;
                            context$3$0.next = 10;
                            return transform(item, index);

                        case 10:
                            index++;

                        case 11:
                            _iteratorNormalCompletion = true;
                            context$3$0.next = 6;
                            break;

                        case 14:
                            context$3$0.next = 20;
                            break;

                        case 16:
                            context$3$0.prev = 16;
                            context$3$0.t0 = context$3$0["catch"](4);
                            _didIteratorError = true;
                            _iteratorError = context$3$0.t0;

                        case 20:
                            context$3$0.prev = 20;
                            context$3$0.prev = 21;

                            if (!_iteratorNormalCompletion && _iterator["return"]) {
                                _iterator["return"]();
                            }

                        case 23:
                            context$3$0.prev = 23;

                            if (!_didIteratorError) {
                                context$3$0.next = 26;
                                break;
                            }

                            throw _iteratorError;

                        case 26:
                            return context$3$0.finish(23);

                        case 27:
                            return context$3$0.finish(20);

                        case 28:
                        case "end":
                            return context$3$0.stop();
                    }
                }, generator, this, [[4, 16, 20, 28], [21,, 23, 27]]);
            });

            return this._wrap(generator);
        }

        /**
         * Applies the specified transformations to all elements of the
         * Sequence. The first transformation returns an iterable. The second
         * one, if present, is called for each element of the output of the
         * first, and returns an arbitrary value. The result is either a concatenation
         * of the iterables returned by the first transformation, or a sequence
         * containing the values returned by the second.
         * 
         * @param {function(item: *, index: number): Iterable} getIterable A
         *     function that returns an iterable for each sequence element. The 
         *     first argument, `item`, is the current sequence element, the
         *     second one, `index`, is the zero-based index of the current element.
         * 
         * @param {function(item: *, innerItem: *): *} [transform] A function
         *     that is called for each element of the iterables returned by
         *     `getIterable`. The final sequence contains the output of this
         *     function. The first argument, `item`, is the current item of the original
         *     sequence, the second, `innerItem`, is the current element of the iterable
         *     returned by `getIterable`.
         * 
         * @return {Sequence} A sequence of the values returned by the composition
         *     of the transformation functions.
         * 
         * @throws {Error} If the object returned by `getIterable` is not iterable.
         * 
         * @example
         * // Simple example, only the `getIterable` function is used
         * const taskLists = [
         *     {tasks: [1]}, {tasks: [2, 3]}, {tasks: [4]},
         *     {tasks: []}, {tasks: [5]}
         * ];
         * 
         * const allTasks = from(taskLists).selectMany(t => t.tasks);
         * 
         * for (let t of allTasks) {
         *     console.log(t); // 1 2 3 4 5
         * }
         * 
         * // Here we use both transformation functions
         * const tasksByDate = [
         *     {
         *         date: "2015-09-20",
         *         tasks: [
         *             { id: 1, text: "buy groceries" },
         *             { id: 2, text: "do laundry" },
         *             { id: 3, text: "meet Janet" }
         *         ]
         *     },
         *     // ... more objects like above ...
         * ];
         * 
         * const allTasks =
         *     from(tasksByDate)
         *         .selectMany(date => date.tasks, (date, task) => task.text);
         * 
         * for (let task of allTasks) {
         *     console.log(task);
         * }
         * 
         * // Output:
         * //     buy groceries
         * //     do laundry
         * //     meet Janet
         * //     ...
         */
    }, {
        key: "selectMany",
        value: function selectMany(getIterable) {
            var transform = arguments.length <= 1 || arguments[1] === undefined ? function (_, n) {
                return n;
            } : arguments[1];

            ensureIsFunction(getIterable, "`getIterable` should be a function");
            ensureIsFunction(transform, "`transform` should be a function");

            var generator = _regeneratorRuntime.mark(function generator() {
                var index, _iteratorNormalCompletion2, _didIteratorError2, _iteratorError2, _iterator2, _step2, item, innerIterable, _iteratorNormalCompletion3, _didIteratorError3, _iteratorError3, _iterator3, _step3, innerItem;

                return _regeneratorRuntime.wrap(function generator$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            index = 0;
                            _iteratorNormalCompletion2 = true;
                            _didIteratorError2 = false;
                            _iteratorError2 = undefined;
                            context$3$0.prev = 4;
                            _iterator2 = _getIterator(this.iterable);

                        case 6:
                            if (_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done) {
                                context$3$0.next = 40;
                                break;
                            }

                            item = _step2.value;
                            innerIterable = getIterable(item, index);

                            ensureIsIterable(innerIterable, "`getIterable` should return an iterable");

                            _iteratorNormalCompletion3 = true;
                            _didIteratorError3 = false;
                            _iteratorError3 = undefined;
                            context$3$0.prev = 13;
                            _iterator3 = _getIterator(innerIterable);

                        case 15:
                            if (_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done) {
                                context$3$0.next = 22;
                                break;
                            }

                            innerItem = _step3.value;
                            context$3$0.next = 19;
                            return transform(item, innerItem);

                        case 19:
                            _iteratorNormalCompletion3 = true;
                            context$3$0.next = 15;
                            break;

                        case 22:
                            context$3$0.next = 28;
                            break;

                        case 24:
                            context$3$0.prev = 24;
                            context$3$0.t0 = context$3$0["catch"](13);
                            _didIteratorError3 = true;
                            _iteratorError3 = context$3$0.t0;

                        case 28:
                            context$3$0.prev = 28;
                            context$3$0.prev = 29;

                            if (!_iteratorNormalCompletion3 && _iterator3["return"]) {
                                _iterator3["return"]();
                            }

                        case 31:
                            context$3$0.prev = 31;

                            if (!_didIteratorError3) {
                                context$3$0.next = 34;
                                break;
                            }

                            throw _iteratorError3;

                        case 34:
                            return context$3$0.finish(31);

                        case 35:
                            return context$3$0.finish(28);

                        case 36:

                            index++;

                        case 37:
                            _iteratorNormalCompletion2 = true;
                            context$3$0.next = 6;
                            break;

                        case 40:
                            context$3$0.next = 46;
                            break;

                        case 42:
                            context$3$0.prev = 42;
                            context$3$0.t1 = context$3$0["catch"](4);
                            _didIteratorError2 = true;
                            _iteratorError2 = context$3$0.t1;

                        case 46:
                            context$3$0.prev = 46;
                            context$3$0.prev = 47;

                            if (!_iteratorNormalCompletion2 && _iterator2["return"]) {
                                _iterator2["return"]();
                            }

                        case 49:
                            context$3$0.prev = 49;

                            if (!_didIteratorError2) {
                                context$3$0.next = 52;
                                break;
                            }

                            throw _iteratorError2;

                        case 52:
                            return context$3$0.finish(49);

                        case 53:
                            return context$3$0.finish(46);

                        case 54:
                        case "end":
                            return context$3$0.stop();
                    }
                }, generator, this, [[4, 42, 46, 54], [13, 24, 28, 36], [29,, 31, 35], [47,, 49, 53]]);
            });

            return this._wrap(generator);
        }

        /**
         * Filters the Sequence by a condition specified.
         * 
         * @param {function(item: *, index: number): boolean} matches A
         *     function that returns true if an element is to be included in 
         *     the result. The first argument, `item`, is the current sequence
         *     element, the second one, `index`, is the zero-based index of the
         *     current element.
         *  
         * @return {Sequence} A Sequence of the elements for which the
         *     'matches' function returned true.
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @example
         * // Simple case, only the `item` parameter of `matches` is used
         * const numbers = [1, 2, 3, 4, 5],
         *       even = from(numbers).where(n => n % 2 === 0);
         *  
         * for (let n of even) console.log(n); // 2 4
         * 
         * // Here we also use the `index` parameter
         * const numbers = [2, 1, 3, 0, 4],
         *       itemEqualsIndex = (item, index) => item === index,
         *       numbersEqualToTheirIndices = from(numbers).where(itemEqualsIndex);
         * 
         * for (let n of numbersEqualToTheirIndices) {
         *     console.log(n); // 1 4
         * }
         */
    }, {
        key: "where",
        value: function where(matches) {
            ensureIsFunction(matches, "`matches` should be a function");

            var generator = _regeneratorRuntime.mark(function generator() {
                var index, _iteratorNormalCompletion4, _didIteratorError4, _iteratorError4, _iterator4, _step4, item;

                return _regeneratorRuntime.wrap(function generator$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            index = 0;
                            _iteratorNormalCompletion4 = true;
                            _didIteratorError4 = false;
                            _iteratorError4 = undefined;
                            context$3$0.prev = 4;
                            _iterator4 = _getIterator(this.iterable);

                        case 6:
                            if (_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done) {
                                context$3$0.next = 15;
                                break;
                            }

                            item = _step4.value;

                            if (!matches(item, index)) {
                                context$3$0.next = 11;
                                break;
                            }

                            context$3$0.next = 11;
                            return item;

                        case 11:
                            index++;

                        case 12:
                            _iteratorNormalCompletion4 = true;
                            context$3$0.next = 6;
                            break;

                        case 15:
                            context$3$0.next = 21;
                            break;

                        case 17:
                            context$3$0.prev = 17;
                            context$3$0.t0 = context$3$0["catch"](4);
                            _didIteratorError4 = true;
                            _iteratorError4 = context$3$0.t0;

                        case 21:
                            context$3$0.prev = 21;
                            context$3$0.prev = 22;

                            if (!_iteratorNormalCompletion4 && _iterator4["return"]) {
                                _iterator4["return"]();
                            }

                        case 24:
                            context$3$0.prev = 24;

                            if (!_didIteratorError4) {
                                context$3$0.next = 27;
                                break;
                            }

                            throw _iteratorError4;

                        case 27:
                            return context$3$0.finish(24);

                        case 28:
                            return context$3$0.finish(21);

                        case 29:
                        case "end":
                            return context$3$0.stop();
                    }
                }, generator, this, [[4, 17, 21, 29], [22,, 24, 28]]);
            });

            return this._wrap(generator);
        }

        /*
         * Join methods
         */

        /*
         * Set methods
         */

        /**
         * Determines whether all elements satisfy a condition.
         * 
         * @param {function(i: *): boolean} matches A function that
         *     determines whether an element of the Sequence satisfies
         *     a condition.
         * 
         * @return {boolean} true if all elements satisfy the condition,
         *     otherwise, false.
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @example
         * const numbers = [1, 2, 3],
         *       even = [2, 4, 6],
         *       isEven = n => n % 2 === 0;
         * 
         * from(numbers).all(isEven); // false
         * from(even).all(isEven); // true
         */
    }, {
        key: "all",
        value: function all(matches) {
            ensureIsFunction(matches, "`matches` should be a function");

            var _iteratorNormalCompletion5 = true;
            var _didIteratorError5 = false;
            var _iteratorError5 = undefined;

            try {
                for (var _iterator5 = _getIterator(this.iterable), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
                    var i = _step5.value;

                    if (!matches(i)) {
                        return false;
                    }
                }
            } catch (err) {
                _didIteratorError5 = true;
                _iteratorError5 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion5 && _iterator5["return"]) {
                        _iterator5["return"]();
                    }
                } finally {
                    if (_didIteratorError5) {
                        throw _iteratorError5;
                    }
                }
            }

            return true;
        }

        /**
         * Determines whether at least one element satisfies a condition.
         * 
         * Without a condition, `!c.any()` can be used to quickly test whether
         * a sequence is empty. This is the preferred way of testing emptiness
         * vs. `c.count() === 0`, as it always runs in O(1) time.
         * 
         * @param {function(i: *): boolean} matches A function that
         *     determines whether an element of the Sequence satisfies
         *     a condition.
         * 
         * @return {boolean} true if at least one element satisfies the
         *     condition, otherwise, false.
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @example
         * // Without a condition, `any` can be used to quickly test whether
         * // a sequence is empty.
         * const empty = [],
         *       nonEmpty = [1];
         * 
         * console.log(from(empty).any()); // false
         * console.log(from(nonEmpty).any()); // true
         * 
         * // With a condition
         * const numbers = [1, 2, 3],
         *       even = [2, 4, 6],
         *       isOdd = n => n % 2 !== 0;
         * from(numbers).any(isOdd); // true
         * from(even).any(isOdd); // false
         */
    }, {
        key: "any",
        value: function any() {
            var matches = arguments.length <= 0 || arguments[0] === undefined ? function (_) {
                return true;
            } : arguments[0];

            ensureIsFunction(matches, "`matches` should be a function");

            var _iteratorNormalCompletion6 = true;
            var _didIteratorError6 = false;
            var _iteratorError6 = undefined;

            try {
                for (var _iterator6 = _getIterator(this.iterable), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
                    var i = _step6.value;

                    if (matches(i)) {
                        return true;
                    }
                }
            } catch (err) {
                _didIteratorError6 = true;
                _iteratorError6 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion6 && _iterator6["return"]) {
                        _iterator6["return"]();
                    }
                } finally {
                    if (_didIteratorError6) {
                        throw _iteratorError6;
                    }
                }
            }

            return false;
        }

        /**
         * Determines whether at least one element is equal to the item specified
         * using the strict equality operator.
         * 
         * @param {*} item The item to find.
         * @return {boolean} true if at least one element is equal to the item
         *     specified, otherwise, false.
         * 
         * @example
         * const numbers = [1, 2, 3];
         * from(numbers).contains(n => n % 2 === 0); // true, 2 is even
         * from(numbers).contains(n => n > 3); // false, all elements less than 3
         */
    }, {
        key: "contains",
        value: function contains(item) {
            var _iteratorNormalCompletion7 = true;
            var _didIteratorError7 = false;
            var _iteratorError7 = undefined;

            try {
                for (var _iterator7 = _getIterator(this.iterable), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
                    var i = _step7.value;
                    if (i === item) return true;
                }
            } catch (err) {
                _didIteratorError7 = true;
                _iteratorError7 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion7 && _iterator7["return"]) {
                        _iterator7["return"]();
                    }
                } finally {
                    if (_didIteratorError7) {
                        throw _iteratorError7;
                    }
                }
            }

            return false;
        }

        /**
         * Concatenates the iterable specified with the current one.
         * 
         * @param {Iterable} other The iterable to concatenate after this instance.
         * @return {Sequence} The concatenation of the two iterables.
         * 
         * @example
         * const numbers = [1, 2],
         *       letters = ["a", "b"],
         *       result  = from(numbers).concat(letters);
         * 
         * for (let i of result) {
         *     console.log(i); // 1 2 a b
         * }
         */
    }, {
        key: "concat",
        value: function concat(other) {
            ensureIsIterable(other, "`other` should be iterable");

            var generator = _regeneratorRuntime.mark(function generator() {
                var _iteratorNormalCompletion8, _didIteratorError8, _iteratorError8, _iterator8, _step8, i, _iteratorNormalCompletion9, _didIteratorError9, _iteratorError9, _iterator9, _step9, j;

                return _regeneratorRuntime.wrap(function generator$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            _iteratorNormalCompletion8 = true;
                            _didIteratorError8 = false;
                            _iteratorError8 = undefined;
                            context$3$0.prev = 3;
                            _iterator8 = _getIterator(this.iterable);

                        case 5:
                            if (_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done) {
                                context$3$0.next = 12;
                                break;
                            }

                            i = _step8.value;
                            context$3$0.next = 9;
                            return i;

                        case 9:
                            _iteratorNormalCompletion8 = true;
                            context$3$0.next = 5;
                            break;

                        case 12:
                            context$3$0.next = 18;
                            break;

                        case 14:
                            context$3$0.prev = 14;
                            context$3$0.t0 = context$3$0["catch"](3);
                            _didIteratorError8 = true;
                            _iteratorError8 = context$3$0.t0;

                        case 18:
                            context$3$0.prev = 18;
                            context$3$0.prev = 19;

                            if (!_iteratorNormalCompletion8 && _iterator8["return"]) {
                                _iterator8["return"]();
                            }

                        case 21:
                            context$3$0.prev = 21;

                            if (!_didIteratorError8) {
                                context$3$0.next = 24;
                                break;
                            }

                            throw _iteratorError8;

                        case 24:
                            return context$3$0.finish(21);

                        case 25:
                            return context$3$0.finish(18);

                        case 26:
                            _iteratorNormalCompletion9 = true;
                            _didIteratorError9 = false;
                            _iteratorError9 = undefined;
                            context$3$0.prev = 29;
                            _iterator9 = _getIterator(other);

                        case 31:
                            if (_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done) {
                                context$3$0.next = 38;
                                break;
                            }

                            j = _step9.value;
                            context$3$0.next = 35;
                            return j;

                        case 35:
                            _iteratorNormalCompletion9 = true;
                            context$3$0.next = 31;
                            break;

                        case 38:
                            context$3$0.next = 44;
                            break;

                        case 40:
                            context$3$0.prev = 40;
                            context$3$0.t1 = context$3$0["catch"](29);
                            _didIteratorError9 = true;
                            _iteratorError9 = context$3$0.t1;

                        case 44:
                            context$3$0.prev = 44;
                            context$3$0.prev = 45;

                            if (!_iteratorNormalCompletion9 && _iterator9["return"]) {
                                _iterator9["return"]();
                            }

                        case 47:
                            context$3$0.prev = 47;

                            if (!_didIteratorError9) {
                                context$3$0.next = 50;
                                break;
                            }

                            throw _iteratorError9;

                        case 50:
                            return context$3$0.finish(47);

                        case 51:
                            return context$3$0.finish(44);

                        case 52:
                        case "end":
                            return context$3$0.stop();
                    }
                }, generator, this, [[3, 14, 18, 26], [19,, 21, 25], [29, 40, 44, 52], [45,, 47, 51]]);
            });

            return this._wrap(generator);
        }

        /**
         * Returns the distinct elements in the sequence (removes duplication).
         * 
         * @return {Sequence} A Sequence containing the distinct elements
         *     of the original one.
         * 
         * @example
         * const numbers = [1, 1, 1, 2, 3, 4, 3],
         *       noDupl = from(numbers).distinct();
         * for (let n of noDupl) console.log(n); // 1 2 3 4
         */
    }, {
        key: "distinct",
        value: function distinct() {
            return new Sequence(new _Set(this.iterable));
        }

        /**
         * Returns all elements of the iterable except the ones in `other`.
         * 
         * @param {Iterable} other The iterable to be subtracted.
         * @return {Sequence} All elements of the iterable except the ones
         *     in `other`.
         * 
         * @example
         * const numbers = [1, 2, 3, 4, 5],
         *       exceptions = [3, 4],
         *       difference = from(numbers).except(exceptions);
         * for (let n of difference) console.log(n); // 1 2 5
         */
    }, {
        key: "except",
        value: function except(other) {
            var iterable = this.iterable;
            other = new Sequence(other);
            return this._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                var _iteratorNormalCompletion10, _didIteratorError10, _iteratorError10, _iterator10, _step10, i;

                return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            _iteratorNormalCompletion10 = true;
                            _didIteratorError10 = false;
                            _iteratorError10 = undefined;
                            context$3$0.prev = 3;
                            _iterator10 = _getIterator(iterable);

                        case 5:
                            if (_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done) {
                                context$3$0.next = 13;
                                break;
                            }

                            i = _step10.value;

                            if (other.contains(i)) {
                                context$3$0.next = 10;
                                break;
                            }

                            context$3$0.next = 10;
                            return i;

                        case 10:
                            _iteratorNormalCompletion10 = true;
                            context$3$0.next = 5;
                            break;

                        case 13:
                            context$3$0.next = 19;
                            break;

                        case 15:
                            context$3$0.prev = 15;
                            context$3$0.t0 = context$3$0["catch"](3);
                            _didIteratorError10 = true;
                            _iteratorError10 = context$3$0.t0;

                        case 19:
                            context$3$0.prev = 19;
                            context$3$0.prev = 20;

                            if (!_iteratorNormalCompletion10 && _iterator10["return"]) {
                                _iterator10["return"]();
                            }

                        case 22:
                            context$3$0.prev = 22;

                            if (!_didIteratorError10) {
                                context$3$0.next = 25;
                                break;
                            }

                            throw _iteratorError10;

                        case 25:
                            return context$3$0.finish(22);

                        case 26:
                            return context$3$0.finish(19);

                        case 27:
                        case "end":
                            return context$3$0.stop();
                    }
                }, callee$2$0, this, [[3, 15, 19, 27], [20,, 22, 26]]);
            }));
        }

        /**
         * Returns the elements both iterables (this and `other`) contain.
         * 
         * @param {Iterable} other The iterable to intersect the current one with.
         * @return {Sequence} The elements both iterables contain.
         * 
         * @example
         * const a = [1, 2, 3],
         *       b = [2, 3, 4],
         *       i = from(a).intersect(b);
         * for (let n of i) console.log(i); // 2 3
         */
    }, {
        key: "intersect",
        value: function intersect(other) {
            var iterable = this.iterable;
            other = new Sequence(other);
            return this._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                var _iteratorNormalCompletion11, _didIteratorError11, _iteratorError11, _iterator11, _step11, i;

                return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            _iteratorNormalCompletion11 = true;
                            _didIteratorError11 = false;
                            _iteratorError11 = undefined;
                            context$3$0.prev = 3;
                            _iterator11 = _getIterator(iterable);

                        case 5:
                            if (_iteratorNormalCompletion11 = (_step11 = _iterator11.next()).done) {
                                context$3$0.next = 13;
                                break;
                            }

                            i = _step11.value;

                            if (!other.contains(i)) {
                                context$3$0.next = 10;
                                break;
                            }

                            context$3$0.next = 10;
                            return i;

                        case 10:
                            _iteratorNormalCompletion11 = true;
                            context$3$0.next = 5;
                            break;

                        case 13:
                            context$3$0.next = 19;
                            break;

                        case 15:
                            context$3$0.prev = 15;
                            context$3$0.t0 = context$3$0["catch"](3);
                            _didIteratorError11 = true;
                            _iteratorError11 = context$3$0.t0;

                        case 19:
                            context$3$0.prev = 19;
                            context$3$0.prev = 20;

                            if (!_iteratorNormalCompletion11 && _iterator11["return"]) {
                                _iterator11["return"]();
                            }

                        case 22:
                            context$3$0.prev = 22;

                            if (!_didIteratorError11) {
                                context$3$0.next = 25;
                                break;
                            }

                            throw _iteratorError11;

                        case 25:
                            return context$3$0.finish(22);

                        case 26:
                            return context$3$0.finish(19);

                        case 27:
                        case "end":
                            return context$3$0.stop();
                    }
                }, callee$2$0, this, [[3, 15, 19, 27], [20,, 22, 26]]);
            }));
        }

        /**
         * Returns the distinct elements from both iterables (this and `other`).
         * 
         * @param {Iterable} other The iterable to union the current one with.
         * @return {Sequence} The distinct elements from both iterables.
         * 
         * @example
         * const a = [1, 2, 1, 3, 2],
         *       b = [3, 3, 4, 5, 4],
         *       u = from(a).union(b);
         * for (let i of u) console.log(i); // 1 2 3 4 5
         */
    }, {
        key: "union",
        value: function union(other) {
            return new Sequence(this.iterable).concat(other).distinct();
        }

        /*
         * Ordering methods
         */

    }, {
        key: "orderBy",
        value: function orderBy(get) {
            var compare = arguments.length <= 1 || arguments[1] === undefined ? compareDefault : arguments[1];

            var result = _Array$from(this.iterable);
            result.sort(function (a, b) {
                return compare(get(a), get(b));
            });
            return new Sequence(result);
        }
    }, {
        key: "orderByDescending",
        value: function orderByDescending(get) {
            var compare = arguments.length <= 1 || arguments[1] === undefined ? compareDefault : arguments[1];

            var result = _Array$from(this.iterable);
            result.sort(function (a, b) {
                return compare(get(b), get(a));
            });
            return new Sequence(result);
        }
    }, {
        key: "reverse",
        value: function reverse() {
            var array = _Array$from(this.iterable);
            return this._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                var i;
                return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            i = array.length - 1;

                        case 1:
                            if (!(i >= 0)) {
                                context$3$0.next = 7;
                                break;
                            }

                            context$3$0.next = 4;
                            return array[i];

                        case 4:
                            i--;
                            context$3$0.next = 1;
                            break;

                        case 7:
                        case "end":
                            return context$3$0.stop();
                    }
                }, callee$2$0, this);
            }));
        }

        /*
         * Grouping methods
         */

    }, {
        key: "groupBy",
        value: function groupBy(getKey) {
            var map = new _Map();
            var _iteratorNormalCompletion12 = true;
            var _didIteratorError12 = false;
            var _iteratorError12 = undefined;

            try {
                for (var _iterator12 = _getIterator(this.iterable), _step12; !(_iteratorNormalCompletion12 = (_step12 = _iterator12.next()).done); _iteratorNormalCompletion12 = true) {
                    var i = _step12.value;

                    var key = getKey(i);
                    if (map.has(key)) map.get(key).push(i);else map.set(getKey(i), [i]);
                }
            } catch (err) {
                _didIteratorError12 = true;
                _iteratorError12 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion12 && _iterator12["return"]) {
                        _iterator12["return"]();
                    }
                } finally {
                    if (_didIteratorError12) {
                        throw _iteratorError12;
                    }
                }
            }

            return new Sequence(map);
        }

        /*
         * Aggregate methods
         */

    }, {
        key: "aggregate",
        value: function aggregate(aggr) {
            var acc = undefined;
            var _iteratorNormalCompletion13 = true;
            var _didIteratorError13 = false;
            var _iteratorError13 = undefined;

            try {
                for (var _iterator13 = _getIterator(this.iterable), _step13; !(_iteratorNormalCompletion13 = (_step13 = _iterator13.next()).done); _iteratorNormalCompletion13 = true) {
                    var i = _step13.value;
                    acc = !acc ? i : aggr(acc, i);
                }
            } catch (err) {
                _didIteratorError13 = true;
                _iteratorError13 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion13 && _iterator13["return"]) {
                        _iterator13["return"]();
                    }
                } finally {
                    if (_didIteratorError13) {
                        throw _iteratorError13;
                    }
                }
            }

            return acc;
        }
    }, {
        key: "average",
        value: function average() {
            var select = arguments.length <= 0 || arguments[0] === undefined ? function (n) {
                return n;
            } : arguments[0];

            var sum = 0,
                count = 0;
            var _iteratorNormalCompletion14 = true;
            var _didIteratorError14 = false;
            var _iteratorError14 = undefined;

            try {
                for (var _iterator14 = _getIterator(this.iterable), _step14; !(_iteratorNormalCompletion14 = (_step14 = _iterator14.next()).done); _iteratorNormalCompletion14 = true) {
                    var i = _step14.value;

                    ensureIsNumber(i, "Only numbers can be averaged");
                    sum += select(i);
                    count++;
                }
            } catch (err) {
                _didIteratorError14 = true;
                _iteratorError14 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion14 && _iterator14["return"]) {
                        _iterator14["return"]();
                    }
                } finally {
                    if (_didIteratorError14) {
                        throw _iteratorError14;
                    }
                }
            }

            return sum / count;
        }
    }, {
        key: "sum",
        value: function sum() {
            var select = arguments.length <= 0 || arguments[0] === undefined ? function (n) {
                return n;
            } : arguments[0];

            var sum = undefined;
            var _iteratorNormalCompletion15 = true;
            var _didIteratorError15 = false;
            var _iteratorError15 = undefined;

            try {
                for (var _iterator15 = _getIterator(this.iterable), _step15; !(_iteratorNormalCompletion15 = (_step15 = _iterator15.next()).done); _iteratorNormalCompletion15 = true) {
                    var i = _step15.value;

                    ensureIsNumber(i, "Only numbers can be summed");
                    sum += select(i);
                }
            } catch (err) {
                _didIteratorError15 = true;
                _iteratorError15 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion15 && _iterator15["return"]) {
                        _iterator15["return"]();
                    }
                } finally {
                    if (_didIteratorError15) {
                        throw _iteratorError15;
                    }
                }
            }

            return sum;
        }

        /**
         * Returns the number of elements in the `Sequence`.
         * 
         * If a `matches` function is specified, returns the number of matching
         * elements.
         * 
         * **Note:** To quickly test whether a sequence is empty, use `!c.any()` instead
         * of `.count() === 0` as the former always runs in O(1) time.
         * 
         * **Complexity:**
         * 
         * - O(1) if the `Iterable` wrapped by this `Sequence` is an `Array`, `Map` or `Set`
         * - O(n) for other `Iterable`s
         * 
         * *Note:* The complexity is O(1) only if the `Array`, `Map` or `Set` is wrapped
         * directly, like `from([1, 2]).count()`. Indirect wrapping will cause the entire
         * sequence to be iterated, as it is the case for
         * `from([1, 2]).where(n => n % 2 === 0).count()`.
         * 
         * @param {function(item: *): boolean} matches A function that should
         *     return `true` for all elements to be included in the count and
         *     `false` for those to be excluded.
         * 
         * @return {number} The number of matching elements in the `Sequence`.
         * 
         * @throws {TypeError} if `matches` is specified, but it is not a function.
         * 
         * @example
         * // No condition specified
         * const numbers = [1, 2, 3, 4],
         *       isEven = n => n % 2 === 0,
         *       count = from(numbers).count();
         * 
         * console.log(count); // 4
         * 
         * // Count matching elements
         * const numbers = [1, 2, 3, 4],
         *       isEven = n => n % 2 === 0,
         *       numberOfEvens = from(numbers).count(isEven);
         * 
         * console.log(numberOfEvens); // 2
         */
    }, {
        key: "count",
        value: function count(matches) {
            if (matches === undefined) {
                // If the wrapped iterable is an `Array`, `Map` or `Set`,
                // we can use these O(1) shortcuts to get the length.
                if (this.iterable instanceof Array) {
                    return this.iterable.length;
                } else if (this.iterable instanceof _Set || this.iterable instanceof _Map) {
                    return this.iterable.size;
                }
                matches = function (_) {
                    return true;
                };
            } else {
                ensureIsFunction(matches, "`matches` should be a function");
            }

            var count = 0;

            var _iteratorNormalCompletion16 = true;
            var _didIteratorError16 = false;
            var _iteratorError16 = undefined;

            try {
                for (var _iterator16 = _getIterator(this.iterable), _step16; !(_iteratorNormalCompletion16 = (_step16 = _iterator16.next()).done); _iteratorNormalCompletion16 = true) {
                    var i = _step16.value;

                    if (matches(i)) {
                        count++;
                    }
                }
            } catch (err) {
                _didIteratorError16 = true;
                _iteratorError16 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion16 && _iterator16["return"]) {
                        _iterator16["return"]();
                    }
                } finally {
                    if (_didIteratorError16) {
                        throw _iteratorError16;
                    }
                }
            }

            return count;
        }

        /*
         * Paging methods
         */

    }, {
        key: "elementAt",
        value: function elementAt(index) {
            ensureIsNotNegative(index, "`index` should not be negative");
            var element = this._elementAt(index);
            ensureIsDefined(element, "Index too large");
            return element;
        }
    }, {
        key: "elementAtOrDefault",
        value: function elementAtOrDefault(index, defaultValue) {
            ensureIsNotNegative(index, "`index` should not be negative");
            var element = this._elementAt(index);
            return element !== undefined ? element : defaultValue;
        }
    }, {
        key: "_elementAt",
        value: function _elementAt(index) {
            var i = 0;
            var _iteratorNormalCompletion17 = true;
            var _didIteratorError17 = false;
            var _iteratorError17 = undefined;

            try {
                for (var _iterator17 = _getIterator(this.iterable), _step17; !(_iteratorNormalCompletion17 = (_step17 = _iterator17.next()).done); _iteratorNormalCompletion17 = true) {
                    var item = _step17.value;

                    if (i === index) return i;
                    i++;
                }
            } catch (err) {
                _didIteratorError17 = true;
                _iteratorError17 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion17 && _iterator17["return"]) {
                        _iterator17["return"]();
                    }
                } finally {
                    if (_didIteratorError17) {
                        throw _iteratorError17;
                    }
                }
            }

            return undefined;
        }

        /**
         * Returns the first element of the sequence. If a `matches` function
         * is specified, it returns the first matching element.
         * 
         * **Evaluation:** eager
         * 
         * @param {function(i: *): boolean} [matches] A function that returns
         * `true` if an element satisfies a condition, `false` otherwise.
         * 
         * @return {*} The first (matching) element. 
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @throws {RangeError} if the sequence contains no elements or no
         *     matching element has been found.
         * 
         * @example
         * // No condition specified, simply retrieve the first element of
         * // the sequence:
         * const numbers = [1, 2, 3, 4, 5],
         *       first = from(numbers).first();
         * 
         * console.log(first); // 1
         * 
         * // Getting the first element that matches a condition:
         * const numbers = [1, 2, 3, 4, 5],
         *       firstEven = from(numbers).first(n => n % 2 === 0);
         * 
         * console.log(firstEven); // 2
         */
    }, {
        key: "first",
        value: function first() {
            var matches = arguments.length <= 0 || arguments[0] === undefined ? function (_) {
                return true;
            } : arguments[0];

            ensureIsFunction(matches, "`matches` should be a function");

            var _iteratorNormalCompletion18 = true;
            var _didIteratorError18 = false;
            var _iteratorError18 = undefined;

            try {
                for (var _iterator18 = _getIterator(this.iterable), _step18; !(_iteratorNormalCompletion18 = (_step18 = _iterator18.next()).done); _iteratorNormalCompletion18 = true) {
                    var i = _step18.value;

                    if (matches(i)) {
                        return i;
                    }
                }
            } catch (err) {
                _didIteratorError18 = true;
                _iteratorError18 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion18 && _iterator18["return"]) {
                        _iterator18["return"]();
                    }
                } finally {
                    if (_didIteratorError18) {
                        throw _iteratorError18;
                    }
                }
            }

            throw new RangeError("No matching element found");
        }

        /**
         * Returns the first element of a sequence or a user-specified default value.
         * If a `matches` function is specified, it returns the first matching
         * element or the default value.
         * 
         * **Evaluation:** eager
         * 
         * @param {*} [defaultValue=undefined] The default value to return
         *     if the sequence contains no (matching) elements.
         * 
         * @param {function(i: *): boolean} [matches] A function that returns
         *     `true` if an element satisfies a condition, `false` otherwise.
         * 
         * @return {*} The first (matching) element of the sequence or the default
         *     value.
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @example
         * // Get first element of a non-empty sequence
         * const numbers = [1, 2, 3];
         * console.log(from(numbers).firstOrDefault()); // Output: 1
         * 
         * // Try to get first element of an empty sequence. No default value specified.
         * console.log(from([]).firstOrDefault()); // Output: undefined
         * 
         * // Try to get first element of an empty sequence. 0 is specified as a default.
         * console.log(from([]).firstOrDefault(0)); // Output: 0
         * 
         * // Get first matching element of a sequence with matching elements
         * const numbers = [1, 2, 3, 4];
         * console.log(from(numbers).firstOrDefault(0, n => n % 2 === 0)); // Output: 2
         * 
         * // Get first matching element of a sequence without a matching element
         * const numbers = [1, 3, 5];
         * console.log(from(numbers).firstOrDefault(0, n => n % 2 === 0)); // Output: 0
         */
    }, {
        key: "firstOrDefault",
        value: function firstOrDefault(defaultValue) {
            var matches = arguments.length <= 1 || arguments[1] === undefined ? function (_) {
                return true;
            } : arguments[1];

            ensureIsFunction(matches, "`matches` should be a function");

            var _iteratorNormalCompletion19 = true;
            var _didIteratorError19 = false;
            var _iteratorError19 = undefined;

            try {
                for (var _iterator19 = _getIterator(this.iterable), _step19; !(_iteratorNormalCompletion19 = (_step19 = _iterator19.next()).done); _iteratorNormalCompletion19 = true) {
                    var i = _step19.value;

                    if (matches(i)) {
                        return i;
                    }
                }
            } catch (err) {
                _didIteratorError19 = true;
                _iteratorError19 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion19 && _iterator19["return"]) {
                        _iterator19["return"]();
                    }
                } finally {
                    if (_didIteratorError19) {
                        throw _iteratorError19;
                    }
                }
            }

            return defaultValue;
        }

        /**
         * Returns the last element of the sequence. If a `matches` function
         * is specified, it returns the last matching element.
         * 
         * **Evaluation:** eager
         * 
         * @param {function(i: *): boolean} [matches] A function that returns
         * `true` if an element satisfies a condition, `false` otherwise.
         * 
         * @return {*} The last (matching) element. 
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @throws {RangeError} if the sequence contains no elements or no
         *     matching element has been found.
         * 
         * @example
         * // No condition specified, simply retrieve the last element of
         * // the sequence:
         * const numbers = [1, 2, 3, 4, 5],
         *       last = from(numbers).last();
         * 
         * console.log(last); // 5
         * 
         * // Getting the last element that matches a condition:
         * const numbers = [1, 2, 3, 4, 5],
         *       lastEven = from(numbers).last(n => n % 2 === 0);
         * 
         * console.log(lastEven); // 4
         */
    }, {
        key: "last",
        value: function last() {
            var matches = arguments.length <= 0 || arguments[0] === undefined ? function (_) {
                return true;
            } : arguments[0];

            ensureIsFunction(matches, "`matches` should be a function");

            var _last2 = this._last(matches);

            var found = _last2.found;
            var item = _last2.item;

            if (!found) {
                throw RangeError("No matching element found");
            }
            return item;
        }

        /**
         * Returns the last element of a sequence or a user-specified default value.
         * If a `matches` function is specified, it returns the last matching
         * element or the default value.
         * 
         * **Evaluation:** eager
         * 
         * @param {*} [defaultValue=undefined] The default value to return
         *     if the sequence contains no (matching) elements.
         * 
         * @param {function(i: *): boolean} [matches] A function that returns
         *     `true` if an element satisfies a condition, `false` otherwise.
         * 
         * @return {*} The last (matching) element of the sequence or the default
         *     value.
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @example
         * // Get last element of a non-empty sequence
         * const numbers = [1, 2, 3];
         * console.log(from(numbers).lastOrDefault()); // Output: 3
         * 
         * // Try to get last element of an empty sequence. No default value specified.
         * console.log(from([]).lastOrDefault()); // Output: undefined
         * 
         * // Try to get last element of an empty sequence. 0 is specified as a default.
         * console.log(from([]).lastOrDefault(0)); // Output: 0
         * 
         * // Get last matching element of a sequence with matching elements
         * const numbers = [1, 2, 3, 4];
         * console.log(from(numbers).lastOrDefault(0, n => n % 2 === 0)); // Output: 4
         * 
         * // Get last matching element of a sequence without a matching element
         * const numbers = [1, 3, 5, 7];
         * console.log(from(numbers).lastOrDefault(0, n => n % 2 === 0)); // Output: 0
         */
    }, {
        key: "lastOrDefault",
        value: function lastOrDefault(defaultValue) {
            var matches = arguments.length <= 1 || arguments[1] === undefined ? function (_) {
                return true;
            } : arguments[1];

            ensureIsFunction(matches, "`matches` should be a function");

            var _last3 = this._last(matches);

            var found = _last3.found;
            var item = _last3.item;

            return found ? item : defaultValue;
        }
    }, {
        key: "_last",
        value: function _last() {
            var matches = arguments.length <= 0 || arguments[0] === undefined ? function (_) {
                return true;
            } : arguments[0];

            var found = false,
                item = undefined;

            var _iteratorNormalCompletion20 = true;
            var _didIteratorError20 = false;
            var _iteratorError20 = undefined;

            try {
                for (var _iterator20 = _getIterator(this.iterable), _step20; !(_iteratorNormalCompletion20 = (_step20 = _iterator20.next()).done); _iteratorNormalCompletion20 = true) {
                    var i = _step20.value;

                    if (matches(i)) {
                        item = i;
                        found = true;
                    }
                }
            } catch (err) {
                _didIteratorError20 = true;
                _iteratorError20 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion20 && _iterator20["return"]) {
                        _iterator20["return"]();
                    }
                } finally {
                    if (_didIteratorError20) {
                        throw _iteratorError20;
                    }
                }
            }

            return { found: found, item: item };
        }

        /**
         * Returns the only element of the sequence. If a `matches` function
         * is specified, it returns the single matching element.
         * 
         * **Evaluation:** eager
         * 
         * **Note:** `singleOrDefault` can be used to express that a sequence should
         * contain at most one (matching) element, and the presence of multiple (matching)
         * elements is an error.
         * 
         * @param {function(i: *): boolean} [matches] A function that returns
         * `true` if an element satisfies a condition, `false` otherwise.
         * 
         * @return {*} The single (matching) element. 
         * 
         * @throws {TypeError} if `matches` is not a function
         * 
         * @throws {RangeError} if the sequence contains no elements, if no
         *     matching element has been found, if the sequence contains more
         *     than one (matching) element.
         * 
         * @example
         * // No condition specified, simply retrieve the only element of
         * // the sequence:
         * const numbers = [1],
         *       single = from(numbers).single();
         * 
         * console.log(single); // 1
         * 
         * // Getting the only element that matches a condition:
         * const numbers = [1, 2, 3],
         *       singleEven = from(numbers).single(n => n % 2 === 0);
         * 
         * console.log(singleEven); // 2
         * 
         * // `single` can be used to express that we expect only one matching
         * // element when the presence of more than one matching elements is
         * // a programming error:
         * 
         * // this is expected to contain only one even number
         * const numbers = [1, 2, 3, 4, 5];
         * 
         * // BOOM! this throws a RangeError because the sequence contains more
         * // than one matching element
         * const even = from(numbers).single(even);
         */
    }, {
        key: "single",
        value: function single() {
            var matches = arguments.length <= 0 || arguments[0] === undefined ? function (_) {
                return true;
            } : arguments[0];

            ensureIsFunction(matches, "`matches` should be a function");

            var _single2 = this._single(matches);

            var found = _single2.found;
            var item = _single2.item;

            if (!found) {
                throw RangeError("No matching element found");
            }
            return item;
        }

        /**
         * Returns the single element of a sequence or a user-specified default value.
         * If a `matches` function is specified, it returns the single matching
         * element or the default value.
         * 
         * **Evaluation:** eager
         * 
         * **Note:** `singleOrDefault` can be used to express that a sequence should
         * contain at most one (matching) element, and the presence of multiple (matching)
         * elements is an error.
         * 
         * @param {*} [defaultValue=undefined] The default value to return
         *     if the sequence contains no (matching) elements.
         * 
         * @param {function(i: *): boolean} [matches] A function that returns
         *     `true` if an element satisfies a condition, `false` otherwise.
         * 
         * @return {*} The single (matching) element of the sequence or the default
         *     value.
         * 
         * @throws {TypeError} if `matches` is not a function
         * @throws {RangeError} if the sequence contains more than one matching element
         * 
         * @example
         * // Get the only element of a non-empty sequence
         * const numbers = [1];
         * console.log(from(numbers).singleOrDefault()); // Output: 1
         * 
         * // Try to get the single element of an empty sequence. No default value specified.
         * console.log(from([]).singleOrDefault()); // Output: undefined
         * 
         * // Try to get single element of an empty sequence. 0 is specified as a default.
         * console.log(from([]).singleOrDefault(0)); // Output: 0
         * 
         * // Get single matching element of a sequence with matching elements
         * const numbers = [1, 2, 3];
         * console.log(from(numbers).singleOrDefault(0, n => n % 2 === 0)); // Output: 2
         * 
         * // Get last matching element of a sequence without a matching element
         * const numbers = [1, 3, 5, 7];
         * console.log(from(numbers).singleOrDefault(0, n => n % 2 === 0)); // Output: 0
         * 
         * // Get single matching element of a sequence with multiple matching elements
         * const numbers = [1, 2, 3, 4];
         * 
         * // BOOM! This throws a `RangeError`:
         * console.log(from(numbers).singleOrDefault(0, n => n % 2 === 0));
         */
    }, {
        key: "singleOrDefault",
        value: function singleOrDefault(defaultValue) {
            var matches = arguments.length <= 1 || arguments[1] === undefined ? function (_) {
                return true;
            } : arguments[1];

            ensureIsFunction(matches, "`matches` should be a function");

            var _single3 = this._single(matches);

            var found = _single3.found;
            var item = _single3.item;

            if (!found) {
                return defaultValue;
            }

            return item;
        }
    }, {
        key: "_single",
        value: function _single() {
            var matches = arguments.length <= 0 || arguments[0] === undefined ? function (_) {
                return true;
            } : arguments[0];

            var found = false,
                item = undefined;

            var _iteratorNormalCompletion21 = true;
            var _didIteratorError21 = false;
            var _iteratorError21 = undefined;

            try {
                for (var _iterator21 = _getIterator(this.iterable), _step21; !(_iteratorNormalCompletion21 = (_step21 = _iterator21.next()).done); _iteratorNormalCompletion21 = true) {
                    var i = _step21.value;

                    if (matches(i)) {
                        if (found) {
                            throw RangeError("Sequence contains more than one matching element");
                        }
                        item = i;
                        found = true;
                    }
                }
            } catch (err) {
                _didIteratorError21 = true;
                _iteratorError21 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion21 && _iterator21["return"]) {
                        _iterator21["return"]();
                    }
                } finally {
                    if (_didIteratorError21) {
                        throw _iteratorError21;
                    }
                }
            }

            return { found: found, item: item };
        }

        /**
         * Returns a sequence containing a default value if the input sequence
         * is empty; otherwise returns the input sequence.
         * 
         * **Evaluation:** lazy
         * 
         * @param {*} [defaultValue] if the input sequence is empty, the result
         *     is a sequence containing this value. 
         * 
         * @return {Sequence} a sequence containing a default value if the input
         *     sequence is empty; otherwise the input sequence.
         * 
         * @example
         * // If the input sequence is not empty, the result is the original
         * // sequence
         * const ex1 = from([1, 2]).defaultIfEmpty();
         * 
         * for (let i of ex1) {
         *     console.log(i);
         * }
         * 
         * // Output:
         * //     1
         * //     2
         * 
         * // If the input sequence is empty, the result is a sequence containing
         * // one `undefined` value
         * const ex2 = from([]).defaultIfEmpty();
         * 
         * for (let i of ex2) {
         *     console.log(i);
         * }
         * 
         * // Output:
         * //     undefined
         * 
         * // If the input sequence is empty and a default value is specified,
         * // the result is a sequence the only item of which is the default value
         * const ex3 = from([]).defaultIfEmpty(0);
         * 
         * for (let i of ex2) {
         *     console.log(i);
         * }
         * 
         * // Output:
         * //     0
         */
    }, {
        key: "defaultIfEmpty",
        value: function defaultIfEmpty(defaultValue) {
            var generator = _regeneratorRuntime.mark(function generator() {
                var iterator, next;
                return _regeneratorRuntime.wrap(function generator$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            iterator = _getIterator(this.iterable);
                            next = iterator.next();

                            if (!next.done) {
                                context$3$0.next = 6;
                                break;
                            }

                            context$3$0.next = 5;
                            return defaultValue;

                        case 5:
                            return context$3$0.abrupt("return");

                        case 6:
                            if (next.done) {
                                context$3$0.next = 12;
                                break;
                            }

                            context$3$0.next = 9;
                            return next.value;

                        case 9:
                            next = iterator.next();
                            context$3$0.next = 6;
                            break;

                        case 12:
                        case "end":
                            return context$3$0.stop();
                    }
                }, generator, this);
            });

            return this._wrap(generator);
        }
    }, {
        key: "skip",
        value: function skip(count) {
            var iterable = this.iterable;
            return this._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                var c, _iteratorNormalCompletion22, _didIteratorError22, _iteratorError22, _iterator22, _step22, i;

                return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            c = 0;
                            _iteratorNormalCompletion22 = true;
                            _didIteratorError22 = false;
                            _iteratorError22 = undefined;
                            context$3$0.prev = 4;
                            _iterator22 = _getIterator(iterable);

                        case 6:
                            if (_iteratorNormalCompletion22 = (_step22 = _iterator22.next()).done) {
                                context$3$0.next = 14;
                                break;
                            }

                            i = _step22.value;

                            if (!(++c > count)) {
                                context$3$0.next = 11;
                                break;
                            }

                            context$3$0.next = 11;
                            return i;

                        case 11:
                            _iteratorNormalCompletion22 = true;
                            context$3$0.next = 6;
                            break;

                        case 14:
                            context$3$0.next = 20;
                            break;

                        case 16:
                            context$3$0.prev = 16;
                            context$3$0.t0 = context$3$0["catch"](4);
                            _didIteratorError22 = true;
                            _iteratorError22 = context$3$0.t0;

                        case 20:
                            context$3$0.prev = 20;
                            context$3$0.prev = 21;

                            if (!_iteratorNormalCompletion22 && _iterator22["return"]) {
                                _iterator22["return"]();
                            }

                        case 23:
                            context$3$0.prev = 23;

                            if (!_didIteratorError22) {
                                context$3$0.next = 26;
                                break;
                            }

                            throw _iteratorError22;

                        case 26:
                            return context$3$0.finish(23);

                        case 27:
                            return context$3$0.finish(20);

                        case 28:
                        case "end":
                            return context$3$0.stop();
                    }
                }, callee$2$0, this, [[4, 16, 20, 28], [21,, 23, 27]]);
            }));
        }
    }, {
        key: "skipWhile",
        value: function skipWhile(matches) {
            var iterable = this.iterable;
            return this._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                var wasFalse, _iteratorNormalCompletion23, _didIteratorError23, _iteratorError23, _iterator23, _step23, i;

                return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            wasFalse = false;
                            _iteratorNormalCompletion23 = true;
                            _didIteratorError23 = false;
                            _iteratorError23 = undefined;
                            context$3$0.prev = 4;
                            _iterator23 = _getIterator(iterable);

                        case 6:
                            if (_iteratorNormalCompletion23 = (_step23 = _iterator23.next()).done) {
                                context$3$0.next = 15;
                                break;
                            }

                            i = _step23.value;

                            if (!(wasFalse || !matches(i))) {
                                context$3$0.next = 12;
                                break;
                            }

                            wasFalse = true;
                            context$3$0.next = 12;
                            return i;

                        case 12:
                            _iteratorNormalCompletion23 = true;
                            context$3$0.next = 6;
                            break;

                        case 15:
                            context$3$0.next = 21;
                            break;

                        case 17:
                            context$3$0.prev = 17;
                            context$3$0.t0 = context$3$0["catch"](4);
                            _didIteratorError23 = true;
                            _iteratorError23 = context$3$0.t0;

                        case 21:
                            context$3$0.prev = 21;
                            context$3$0.prev = 22;

                            if (!_iteratorNormalCompletion23 && _iterator23["return"]) {
                                _iterator23["return"]();
                            }

                        case 24:
                            context$3$0.prev = 24;

                            if (!_didIteratorError23) {
                                context$3$0.next = 27;
                                break;
                            }

                            throw _iteratorError23;

                        case 27:
                            return context$3$0.finish(24);

                        case 28:
                            return context$3$0.finish(21);

                        case 29:
                        case "end":
                            return context$3$0.stop();
                    }
                }, callee$2$0, this, [[4, 17, 21, 29], [22,, 24, 28]]);
            }));
        }
    }, {
        key: "take",
        value: function take(count) {
            var iterable = this.iterable;
            return this._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                var c, _iteratorNormalCompletion24, _didIteratorError24, _iteratorError24, _iterator24, _step24, i;

                return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            c = 0;
                            _iteratorNormalCompletion24 = true;
                            _didIteratorError24 = false;
                            _iteratorError24 = undefined;
                            context$3$0.prev = 4;
                            _iterator24 = _getIterator(iterable);

                        case 6:
                            if (_iteratorNormalCompletion24 = (_step24 = _iterator24.next()).done) {
                                context$3$0.next = 17;
                                break;
                            }

                            i = _step24.value;

                            if (!(++c <= count)) {
                                context$3$0.next = 13;
                                break;
                            }

                            context$3$0.next = 11;
                            return i;

                        case 11:
                            context$3$0.next = 14;
                            break;

                        case 13:
                            return context$3$0.abrupt("break", 17);

                        case 14:
                            _iteratorNormalCompletion24 = true;
                            context$3$0.next = 6;
                            break;

                        case 17:
                            context$3$0.next = 23;
                            break;

                        case 19:
                            context$3$0.prev = 19;
                            context$3$0.t0 = context$3$0["catch"](4);
                            _didIteratorError24 = true;
                            _iteratorError24 = context$3$0.t0;

                        case 23:
                            context$3$0.prev = 23;
                            context$3$0.prev = 24;

                            if (!_iteratorNormalCompletion24 && _iterator24["return"]) {
                                _iterator24["return"]();
                            }

                        case 26:
                            context$3$0.prev = 26;

                            if (!_didIteratorError24) {
                                context$3$0.next = 29;
                                break;
                            }

                            throw _iteratorError24;

                        case 29:
                            return context$3$0.finish(26);

                        case 30:
                            return context$3$0.finish(23);

                        case 31:
                        case "end":
                            return context$3$0.stop();
                    }
                }, callee$2$0, this, [[4, 19, 23, 31], [24,, 26, 30]]);
            }));
        }
    }, {
        key: "takeWhile",
        value: function takeWhile(matches) {
            var iterable = this.iterable;
            return this._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                var _iteratorNormalCompletion25, _didIteratorError25, _iteratorError25, _iterator25, _step25, i;

                return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            _iteratorNormalCompletion25 = true;
                            _didIteratorError25 = false;
                            _iteratorError25 = undefined;
                            context$3$0.prev = 3;
                            _iterator25 = _getIterator(iterable);

                        case 5:
                            if (_iteratorNormalCompletion25 = (_step25 = _iterator25.next()).done) {
                                context$3$0.next = 16;
                                break;
                            }

                            i = _step25.value;

                            if (!matches(i)) {
                                context$3$0.next = 12;
                                break;
                            }

                            context$3$0.next = 10;
                            return i;

                        case 10:
                            context$3$0.next = 13;
                            break;

                        case 12:
                            return context$3$0.abrupt("break", 16);

                        case 13:
                            _iteratorNormalCompletion25 = true;
                            context$3$0.next = 5;
                            break;

                        case 16:
                            context$3$0.next = 22;
                            break;

                        case 18:
                            context$3$0.prev = 18;
                            context$3$0.t0 = context$3$0["catch"](3);
                            _didIteratorError25 = true;
                            _iteratorError25 = context$3$0.t0;

                        case 22:
                            context$3$0.prev = 22;
                            context$3$0.prev = 23;

                            if (!_iteratorNormalCompletion25 && _iterator25["return"]) {
                                _iterator25["return"]();
                            }

                        case 25:
                            context$3$0.prev = 25;

                            if (!_didIteratorError25) {
                                context$3$0.next = 28;
                                break;
                            }

                            throw _iteratorError25;

                        case 28:
                            return context$3$0.finish(25);

                        case 29:
                            return context$3$0.finish(22);

                        case 30:
                        case "end":
                            return context$3$0.stop();
                    }
                }, callee$2$0, this, [[3, 18, 22, 30], [23,, 25, 29]]);
            }));
        }

        /*
         * Factory methods
         */

        /**
         * Returns an empty `Sequence`.
         * 
         * **Note:**
         * The `Sequence` instance returned by `empty` is cached, that is,
         * it always returns the same instance.
         * 
         * @return {Sequence} A `Sequence` that has no elements.
         * 
         * @example
         * console.log("start");
         * 
         * for (let i of Sequence.empty()) {
         *     console.log(i);
         * }
         * 
         * console.log("end");
         * 
         * // Output:
         * //     start
         * //     end
         */
    }, {
        key: "toArray",

        /*
         * Conversion methods
         */

        value: function toArray() {
            return _Array$from(this.iterable);
        }

        /*
         * Private helpers
         */

    }, {
        key: "_wrap",
        value: function _wrap(generator) {
            return new Sequence(_defineProperty({}, _Symbol$iterator, generator.bind(this)));
        }
    }, {
        key: "_log",
        value: function _log() {
            var _iteratorNormalCompletion26 = true;
            var _didIteratorError26 = false;
            var _iteratorError26 = undefined;

            try {
                for (var _iterator26 = _getIterator(this.iterable), _step26; !(_iteratorNormalCompletion26 = (_step26 = _iterator26.next()).done); _iteratorNormalCompletion26 = true) {
                    var i = _step26.value;
                    console.log(i);
                }
            } catch (err) {
                _didIteratorError26 = true;
                _iteratorError26 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion26 && _iterator26["return"]) {
                        _iterator26["return"]();
                    }
                } finally {
                    if (_didIteratorError26) {
                        throw _iteratorError26;
                    }
                }
            }
        }
    }], [{
        key: "empty",
        value: function empty() {
            var self = Sequence;

            if (self._emptyInstance === undefined) {
                self._emptyInstance = self._wrap(_regeneratorRuntime.mark(function callee$2$0() {
                    return _regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
                        while (1) switch (context$3$0.prev = context$3$0.next) {
                            case 0:
                            case "end":
                                return context$3$0.stop();
                        }
                    }, callee$2$0, this);
                }));
            }

            return self._emptyInstance;
        }

        /**
         * Returns a `Sequence` that contains the specified element repeated
         * the specified number of times.
         * 
         * @param {*} item The item to be repeated.
         * @param {number} count How many times to repeat the item.
         * 
         * @return {Sequence} A `Sequence` that contains the specified element
         *     repeated the specified number of times.
         * 
         * @throws {TypeError} if `count` is not a number
         * @throws {RangeError} if `count` is negative
         * 
         * @example
         * for (let i of Sequence.repeat("a", 3)) {
         *     console.log(i);
         * }
         * 
         * // Output:
         * //     a
         * //     a
         * //     a
         */
    }, {
        key: "repeat",
        value: function repeat(item, count) {
            ensureIsNumber(count, "`count` should be a number");
            ensureIsNotNegative(count, "`count` should not be negative");

            var generator = _regeneratorRuntime.mark(function generator() {
                var i;
                return _regeneratorRuntime.wrap(function generator$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            i = 0;

                        case 1:
                            if (!(i < count)) {
                                context$3$0.next = 7;
                                break;
                            }

                            context$3$0.next = 4;
                            return item;

                        case 4:
                            i++;
                            context$3$0.next = 1;
                            break;

                        case 7:
                        case "end":
                            return context$3$0.stop();
                    }
                }, generator, this);
            });

            return Sequence._wrap(generator);
        }

        /**
         * Returns an increasing sequence of integers, of `count` items,
         * starting at `start`.
         * 
         * @param {number} start The first element of the sequence.
         * @param {number} count The number of elements.
         * 
         * @return {Sequence} An increasing sequence of integers, of
         *     `count` items, starting at `start`.
         * 
         * @throws {TypeError} if either `start` or `count` is not a number
         * @throws {RangeError} if count is negative
         * 
         * @example
         * for (let i of Sequence.range(2, 4)) {
         *     console.log(i);
         * }
         * 
         * // Output:
         * //     2
         * //     3
         * //     4
         * //     5
         */
    }, {
        key: "range",
        value: function range(start, count) {
            ensureIsNumber(start, "`start` should be a number");
            ensureIsNumber(count, "`count` should be a number");
            ensureIsNotNegative(count, "`count` should not be negative");

            var generator = _regeneratorRuntime.mark(function generator() {
                var i;
                return _regeneratorRuntime.wrap(function generator$(context$3$0) {
                    while (1) switch (context$3$0.prev = context$3$0.next) {
                        case 0:
                            i = 0;

                        case 1:
                            if (!(i < count)) {
                                context$3$0.next = 7;
                                break;
                            }

                            context$3$0.next = 4;
                            return start + i;

                        case 4:
                            i++;
                            context$3$0.next = 1;
                            break;

                        case 7:
                        case "end":
                            return context$3$0.stop();
                    }
                }, generator, this);
            });

            return Sequence._wrap(generator);
        }
    }, {
        key: "_wrap",
        value: function _wrap(generator) {
            return new Sequence(_defineProperty({}, _Symbol$iterator, generator));
        }
    }]);

    return Sequence;
})();

exports.Sequence = Sequence;
function compareDefault(a, b) {
    if (a < b) return -1;
    if (a > b) return 1;
    return 0;
}

/*
 * Argument validation helpers
 */

function ensureIsNumber(n, message) {
    if (typeof n !== "number") {
        throw new TypeError(message);
    }
}

function ensureIsNotNegative(n, message) {
    if (n < 0) throw new RangeError(message);
}

function ensureIsDefined(x, message) {
    if (x === undefined) throw new TypeError(message);
}

function ensureIsFunction(f, message) {
    if (typeof f !== "function") throw new TypeError(message);
}

function ensureIsIterable(i, message) {
    var iter = i[_Symbol$iterator];
    ensureIsDefined(iter, message);
    ensureIsFunction(iter, message);
}

// TODO: this is a very naive implementation. We should build
// a set from `other` and use that for the lookup.

// TODO: this is a very naive implementation. We should build
// a set from `other` and use that for the lookup.
//# sourceMappingURL=eslinq.js.map
